from hatchling.builders.hooks.plugin.interface import BuildHookInterface
import subprocess


class BuildFrontend(BuildHookInterface):
    PLUGIN_NAME = "build_frontend"
    FRONTEND_DIR_PATH = "vue-components"

    def initialize(self, version, build_data):
        subprocess.run(
            args=["npm", "install"],
            cwd=self.FRONTEND_DIR_PATH,
            check=True,
        )
        subprocess.run(
            args=["npm", "run", "build"],
            cwd=self.FRONTEND_DIR_PATH,
            check=True,
        )

        return super().initialize(version, build_data)



from trame.widgets import plotly, html, client


def result_list_component(state, ctrl, image_ids_key):
    with html.Div(
        v_for=(f"image_id in {image_ids_key}",), key="image_id", style="padding: 0.71875rem;"
    ):
        with client.Getter(name=("image_id + '_result'",), value_name="value"):
            plotly.Figure(
                style="width: 100%; height: 12rem;",
                data=("value",),
            )



from trame.widgets import quasar
from trame.widgets import html


def collapsible_card(collapse_key):
    with quasar.QCard():
        with quasar.QCardSection():
            with html.Div(classes="row items-center no-wrap"):
                title_slot = html.Div(classes="col")

                with html.Div(classes="col-auto"):
                    quasar.QBtn(
                        round=True,
                        flat=True,
                        dense=True,
                        click=f"{collapse_key} = !{collapse_key}",
                        icon=(f"{collapse_key} ? 'keyboard_arrow_down' : 'keyboard_arrow_up'",),
                    )
        with quasar.QSlideTransition():
            with html.Div(v_show=f"!{collapse_key}"):
                content_slot = quasar.QCardSection()
                actions_slot = quasar.QCardActions(align="right")

    return title_slot, content_slot, actions_slot



from nrtk_explorer.app.core import create_engine


def main(server=None, *args, **kwargs):
    engine = create_engine(server)
    engine.server.start(**kwargs)


if __name__ == "__main__":
    main()



from nrtk_explorer.widgets.nrtk_explorer import ScatterPlot
from nrtk_explorer.library import embeddings_extractor
from nrtk_explorer.library import dimension_reducers
from nrtk_explorer.library import images_manager
from nrtk_explorer.app.applet import Applet
import nrtk_explorer.test_data

import asyncio
import json
import os

from trame.widgets import quasar, html
from trame.ui.quasar import QLayout
from trame.app import get_server, asynchronous


os.environ["TRAME_DISABLE_V3_WARNING"] = "1"

DIR_NAME = os.path.dirname(nrtk_explorer.test_data.__file__)
DATASET_DIRS = [
    f"{DIR_NAME}/OIRDS_v1_0/oirds.json",
    f"{DIR_NAME}/OIRDS_v1_0/oirds_test.json",
    f"{DIR_NAME}/OIRDS_v1_0/oirds_train.json",
]


class EmbeddingsApp(Applet):
    def __init__(self, server):
        super().__init__(server)

        self._ui = None
        self._on_select_fn = None
        self.reducer = dimension_reducers.DimReducerManager()
        self.is_standalone_app = self.server.state.parent is None
        if self.is_standalone_app:
            self.context.images_manager = images_manager.ImagesManager()

        if self.state.current_dataset is None:
            self.state.current_dataset = DATASET_DIRS[0]
        self.features = None

        self.state.client_only("camera_position")
        self.state.feature_extraction_model = "resnet50.a1_in1k"

        self.server.controller.add("on_server_ready")(self.on_server_ready)
        self.transformed_images_cache = {}

    def on_server_ready(self, *args, **kwargs):
        # Bind instance methods to state change
        self.on_current_dataset_change()
        self.on_feature_extraction_model_change()
        self.state.change("current_dataset")(self.on_current_dataset_change)
        self.state.change("feature_extraction_model")(self.on_feature_extraction_model_change)

    def on_feature_extraction_model_change(self, **kwargs):
        feature_extraction_model = self.state.feature_extraction_model
        self.extractor = embeddings_extractor.EmbeddingsExtractor(
            model_name=feature_extraction_model, manager=self.context.images_manager
        )

    def on_current_dataset_change(self, **kwargs):
        self.state.num_elements_disabled = True
        with open(self.state.current_dataset) as f:
            dataset = json.load(f)
            self.images = dataset["images"]
            self.state.num_elements_max = len(self.images)
        self.state.num_elements_disabled = False

        if self.is_standalone_app:
            self.context.images_manager = images_manager.ImagesManager()

    def on_run_clicked(self):
        self.state.is_loading = True
        asynchronous.create_task(self.compute(self.compute_source_points))

    async def compute(self, method):
        # We need to yield twice for the is_loading=True to commit to the trame state
        # before this routine ends
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        await method()
        with self.state:
            self.state.is_loading = False

    async def compute_source_points(self):
        self.features = self.extractor.extract(
            paths=self.context.paths,
            batch_size=int(self.state.model_batch_size),
        )

        if self.state.tab == "PCA":
            self.state.points_sources = self.reducer.reduce(
                name="PCA",
                fit_features=self.features,
                features=self.features,
                dims=self.state.dimensionality,
                whiten=self.state.pca_whiten,
                solver=self.state.pca_solver,
            )

        elif self.state.tab == "UMAP":
            args = {}
            if self.state.umap_random_seed:
                args["random_state"] = int(self.state.umap_random_seed_value)

            if self.state.umap_n_neighbors:
                args["n_neighbors"] = int(self.state.umap_n_neighbors_number)

            self.state.points_sources = self.reducer.reduce(
                name="UMAP",
                dims=self.state.dimensionality,
                fit_features=self.features,
                features=self.features,
                **args,
            )

        # Unselect current selection of images
        if self._on_select_fn:
            self._on_select_fn([])

        self.state.points_transformations = []
        self.state.user_selected_points_indices = []
        self.state.camera_position = []

    def on_run_transformations(self, transformed_image_ids):
        # Fillup the cache with the transformed images
        for img_id in transformed_image_ids:
            img = self.context.image_objects[img_id]
            img = self.context.images_manager.prepare_for_model(img)
            self.transformed_images_cache[img_id] = img

        transformation_features = self.extractor.extract(
            paths=transformed_image_ids,
            content=self.transformed_images_cache,
            batch_size=int(self.state.model_batch_size),
        )

        if self.state.tab == "PCA":
            self.state.points_transformations = self.reducer.reduce(
                name="PCA",
                fit_features=self.features,
                features=transformation_features,
                dims=self.state.dimensionality,
                whiten=self.state.pca_whiten,
                solver=self.state.pca_solver,
            )

        elif self.state.tab == "UMAP":
            args = {}
            if self.state.umap_random_seed:
                args["random_state"] = int(self.state.umap_random_seed_value)

            if self.state.umap_n_neighbors:
                args["n_neighbors"] = int(self.state.umap_n_neighbors_number)

            self.state.points_transformations = self.reducer.reduce(
                name="UMAP",
                dims=self.state.dimensionality,
                fit_features=self.features,
                features=transformation_features,
                **args,
            )

    def set_on_select(self, fn):
        self._on_select_fn = fn

    def on_select(self, indices):
        self.state.user_selected_points_indices = indices
        self.state.points_transformations = []
        ids = [self.state.images_ids[i] for i in indices]
        if self._on_select_fn:
            self._on_select_fn(ids)

    def on_move(self, camera_position):
        self.state.camera_position = camera_position

    def set_on_hover(self, fn):
        self._on_hover_fn = fn

    def on_hover(self, point):
        self.state.highlighted_point = point
        image_id = -1
        if point is not None and point in self.state.user_selected_points_indices:
            image_id = self.state.images_ids[int(point)]
        if self._on_hover_fn:
            self._on_hover_fn(image_id)

    def on_image_hovered(self, id_, is_transformation):
        # If the point is in the list of selected points, we set it as the highlighted point
        if id_ in self.state.images_ids:
            index = self.state.images_ids.index(id_)
            if is_transformation:
                index_selected = self.state.user_selected_points_indices.index(index)
                self.state.highlighted_point = len(self.state.points_sources) + index_selected
            else:
                self.state.highlighted_point = index
        else:
            # If the point is not in the list of selected points, we set it to a negative point
            self.state.highlighted_point = -1

    def visualization_widget(self):
        ScatterPlot(
            cameraMove="camera_position=$event",
            cameraPosition=("camera_position",),
            highlightedPoint=("highlighted_point", -1),
            hover=(self.on_hover, "[$event]"),
            points=("points_sources", []),
            transformedPoints=("points_transformations", []),
            select=(self.on_select, "[$event]"),
            selectedPoints=("user_selected_points_indices", []),
        )

    def settings_widget(self):
        with html.Div(trame_server=self.server, classes="col"):
            with html.Div(classes="q-gutter-y-md"):
                quasar.QBtnToggle(
                    v_model=("dimensionality", "3"),
                    toggler_color="primary",
                    flat=True,
                    spread=True,
                    options=(
                        [
                            {"label": "2D", "value": "2"},
                            {"label": "3D", "value": "3"},
                        ],
                    ),
                )

            quasar.QSelect(
                label="Embeddings Model",
                v_model=("feature_extraction_model",),
                options=(
                    [
                        {"label": "ResNet50", "value": "resnet50.a1_in1k"},
                        {"label": "EfficientNet_b0", "value": "efficientnet_b0.ra_in1k"},
                        {
                            "label": "MobileNetV3Large",
                            "value": "mobilenetv3_large_100.ra_in1k",
                        },
                    ],
                ),
                filled=True,
                emit_value=True,
                map_options=True,
            )
            quasar.QInput(
                v_model=("model_batch_size", 32),
                filled=True,
                stack_label=True,
                label="Batch Size",
                type="number",
            )

        with html.Div(classes="col"):
            with quasar.QTabs(
                v_model=("tab", "PCA"),
                dense=True,
                narrow_indicator=True,
                active_color="primary",
                indicator_color="primary",
                align="justify",
            ):
                quasar.QTab(name="PCA", label="pca")
                quasar.QTab(name="UMAP", label="umap")
            quasar.QSeparator()
            with quasar.QTabPanels(v_model=("tab", "PCA")):
                with quasar.QTabPanel(name="PCA"):
                    quasar.QToggle(
                        v_model=("pca_whiten", False),
                        label="Whiten",
                        left_label=True,
                    )
                    quasar.QSelect(
                        v_model=("pca_solver", "auto"),
                        label="SVD Solver",
                        toggler_color="primary",
                        options=(
                            [
                                "auto",
                                "full",
                                "arpack",
                                "randomized",
                            ],
                        ),
                    )

                with quasar.QTabPanel(name="UMAP"):
                    quasar.QToggle(
                        v_model=("umap_n_neighbors", False),
                        label="Number of neighbors",
                        left_label=True,
                    )
                    quasar.QInput(
                        v_model=("umap_n_neighbors_number", 15),
                        disable=("!umap_n_neighbors",),
                        filled=True,
                        stack_label=True,
                        label="Neighbors amount",
                        type="number",
                    )
                    quasar.QToggle(
                        v_model=("umap_random_seed", True),
                        label="Random seed",
                        left_label=True,
                    )
                    quasar.QInput(
                        v_model=("umap_random_seed_value", 1),
                        disable=("!umap_random_seed",),
                        filled=True,
                        stack_label=True,
                        label="Seed value",
                        type="number",
                    )

    def compute_ui(self):
        with html.Div(trame_server=self.server):
            quasar.QBtn(
                label="Compute",
                loading=("is_loading", False),
                click=self.on_run_clicked,
                flat=True,
            )

    # This is only used within when this module (file) is executed as an Standalone app.
    @property
    def ui(self):
        if self._ui is None:
            with QLayout(self.server) as layout:
                with quasar.QHeader():
                    with quasar.QToolbar(classes="shadow-4"):
                        quasar.QToolbarTitle("Embeddings")
                        quasar.QBtn("Reset")

                with quasar.QDrawer(
                    v_model=("leftDrawerOpen", True),
                    side="left",
                    elevated=True,
                ):
                    with html.Div(classes="column justify-center", style="padding:1rem"):
                        with html.Div(classes="col"):
                            quasar.QSeparator()
                            quasar.QSelect(
                                label="Dataset",
                                v_model=("current_dataset",),
                                options=(
                                    [
                                        {"label": "oirds_test", "value": DATASET_DIRS[0]},
                                        {"label": "oirds_train", "value": DATASET_DIRS[1]},
                                    ],
                                ),
                                filled=True,
                                emit_value=True,
                                map_options=True,
                            )
                    self.settings_widget()
                    self.compute_ui()

                # Main content
                with quasar.QPageContainer():
                    with quasar.QPage():
                        with html.Div(classes="row", style="min-height: inherit;"):
                            with html.Div(classes="col q-pa-md"):
                                self.visualization_widget()

                self._ui = layout
        return self._ui


def embeddings(server=None, *args, **kwargs):
    server = get_server()
    server.client_type = "vue3"

    embeddings_app = EmbeddingsApp(server)
    embeddings_app.ui

    server.start(**kwargs)


if __name__ == "__main__":
    embeddings()



import torch
import logging
import numpy as np
import timm

from nrtk_explorer.library import images_manager
from torch.utils.data import DataLoader, Dataset


# Create a dataset for images
class ImagesDataset(Dataset):
    def __init__(self, images):
        self.images = images

    def __len__(self):
        return len(self.images)

    def __getitem__(self, i):
        return self.images[i][0]


class EmbeddingsExtractor:
    def __init__(
        self, model_name="resnet50d", manager=images_manager.ImagesManager(), force_cpu=False
    ):
        self.manager = manager
        self.device = "cuda" if torch.cuda.is_available() and not force_cpu else "cpu"
        self.model = model_name

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        logging.info(f"Using {dev} devices for feature extraction")
        self._device = torch.device(dev)

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model_name):
        # Create model in inference mode
        self._model = (
            timm.create_model(model_name, pretrained=True, num_classes=0).to(self._device).eval()
        )

        # Create transformer for image
        self._model_transformer = timm.data.create_transform(
            **timm.data.resolve_model_data_config(self._model.pretrained_cfg)
        )

    def transform_image(self, img):
        """Transform image to fit model input size and format"""
        return self._model_transformer(img).unsqueeze(0)

    def extract(self, paths, content=None, batch_size=32):
        """Extract features from images in paths"""
        if len(paths) == 0:
            return None

        features = list()
        transformed_images = list()

        # Load images and transform them
        for path in paths:
            img = None
            if content and path in content:
                img = content[path]
            else:
                img = self.manager.load_image_for_model(path)

            transformed_images.append(self.transform_image(img))

        # Extract features from images
        for batch in DataLoader(ImagesDataset(transformed_images), batch_size=batch_size):
            # Copy image to device if using device
            if self.device.type == "cuda":
                batch = batch.cuda()

            features.append(self.model(batch).numpy(force=True))

        return np.vstack(features)



from typing import Any, Optional, Dict, Tuple

from PIL.Image import Image

from nrtk_explorer.library.transforms import ImageTransform, ParameterDescription

from nrtk.impls.perturb_image.generic.cv2.blur import GaussianBlurPerturber
from nrtk.impls.perturb_image.pybsm.perturber import PybsmPerturber, PybsmSensor, PybsmScenario
import pybsm

import numpy as np
from PIL import Image as ImageModule


class NrtkGaussianBlurTransform(ImageTransform):
    def __init__(self, perturber: Optional[GaussianBlurPerturber] = None):
        if perturber is None:
            perturber = GaussianBlurPerturber()

        self._perturber: GaussianBlurPerturber = perturber

    def get_parameters(self) -> Dict[str, Any]:
        return {
            "ksize": self._perturber.ksize,
        }

    def set_parameters(self, params: Dict[str, Any]):
        self._perturber.ksize = params.get("ksize", 1)

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        ksize_description: ParameterDescription = {
            "type": "integer",
            "label": "Kernel Size",
            "default": 1,
            "description": None,
            "options": None,
        }

        return {
            "ksize": ksize_description,
        }

    def execute(self, input: Image, *input_args: Any) -> Image:
        input_array = np.asarray(input)
        output_array = self._perturber.perturb(input_array)

        return ImageModule.fromarray(output_array)


# Taken from the nrtk package tests
def createSampleSensorandScenario() -> Tuple[PybsmSensor, PybsmScenario]:

    name = "L32511x"

    # telescope focal length (m)
    f = 4
    # Telescope diameter (m)
    D = 275e-3

    # detector pitch (m)
    p = 0.008e-3

    # Optical system transmission, red  band first (m)
    optTransWavelengths = np.array([0.58 - 0.08, 0.58 + 0.08]) * 1.0e-6
    # guess at the full system optical transmission (excluding obscuration)
    opticsTransmission = 0.5 * np.ones(optTransWavelengths.shape[0])

    # Relative linear telescope obscuration
    eta = 0.4  # guess

    # detector width is assumed to be equal to the pitch
    wx = p
    wy = p
    # integration time (s) - this is a maximum, the actual integration time will be
    # determined by the well fill percentage
    intTime = 30.0e-3

    # dark current density of 1 nA/cm2 guess, guess mid range for a silicon camera
    darkCurrent = pybsm.darkCurrentFromDensity(1e-5, wx, wy)

    # rms read noise (rms electrons)
    readNoise = 25.0

    # maximum ADC level (electrons)
    maxN = 96000.0

    # bit depth
    bitdepth = 11.9

    # maximum allowable well fill (see the paper for the logic behind this)
    maxWellFill = 0.6

    # jitter (radians) - The Olson paper says that its "good" so we'll guess 1/4 ifov rms
    sx = 0.25 * p / f
    sy = sx

    # drift (radians/s) - again, we'll guess that it's really good
    dax = 100e-6
    day = dax

    # etector quantum efficiency as a function of wavelength (microns)
    # for a generic high quality back-illuminated silicon array
    # https://www.photometrics.com/resources/learningzone/quantumefficiency.php
    qewavelengths = np.array([0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1]) * 1.0e-6
    qe = np.array([0.05, 0.6, 0.75, 0.85, 0.85, 0.75, 0.5, 0.2, 0])

    sensor = PybsmSensor(
        name,
        D,
        f,
        p,
        optTransWavelengths,
        opticsTransmission,
        eta,
        wx,
        wy,
        intTime,
        darkCurrent,
        readNoise,
        maxN,
        bitdepth,
        maxWellFill,
        sx,
        sy,
        dax,
        day,
        qewavelengths,
        qe,
    )

    altitude = 9000.0
    # range to target
    groundRange = 60000.0

    scenario_name = "niceday"
    # weather model
    ihaze = 1
    scenario = PybsmScenario(scenario_name, ihaze, altitude, groundRange)
    scenario.aircraftSpeed = 100.0

    return sensor, scenario


class NrtkPybsmTransform(ImageTransform):
    def __init__(self, perturber: Optional[PybsmPerturber] = None):
        if perturber is None:
            sensor, scenario = createSampleSensorandScenario()
            perturber = PybsmPerturber(sensor=sensor, scenario=scenario)

        self._perturber: PybsmPerturber = perturber

    def get_parameters(self) -> dict[str, Any]:
        return {
            "D": self._perturber.sensor.D,
            "f": self._perturber.sensor.f,
        }

    def set_parameters(self, params: Dict[str, Any]):
        self._perturber.sensor.D = params["D"]
        self._perturber.sensor.f = params["f"]
        self._perturber.metrics = pybsm.niirs(self._perturber.sensor, self._perturber.scenario)

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        aperture_description: ParameterDescription = {
            "type": "float",
            "label": "Effective Aperture (m)",
            "default": None,
            "description": None,
            "options": None,
        }

        focal_description: ParameterDescription = {
            "type": "float",
            "label": "Focal Length (m)",
            "default": None,
            "description": None,
            "options": None,
        }

        return {
            "D": aperture_description,
            "f": focal_description,
        }

    def execute(self, input: Image, *input_args: Any) -> Image:
        if len(input_args) == 0:
            input_args = ({"img_gsd": 0.15},)

        input_array = np.asarray(input)
        output_array = self._perturber.perturb(input_array, *input_args)

        return ImageModule.fromarray(output_array)



from trame_client.widgets.core import AbstractElement
from .. import module


class HtmlElement(AbstractElement):
    def __init__(self, _elem_name, children=None, **kwargs):
        super().__init__(_elem_name, children, **kwargs)
        if self.server:
            self.server.enable_module(module)


class ImageDetection(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "image-detection",
            **kwargs,
        )
        self._attr_names += [
            "identifier",
            "src",
            "meta",
            "annotations",
            "categories",
            "selected",
            "isTransformation",
        ]
        self._event_names += [
            "hover",
        ]


class ScatterPlot(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "scatter-plot",
            **kwargs,
        )
        self._attr_names += [
            "cameraPosition",
            "highlightedPoint",
            "points",
            "transformedPoints",
            "selectedPoints",
        ]
        self._event_names += [
            "cameraMove",
            "hover",
            "select",
        ]


class ParamsWidget(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "params-widget",
            **kwargs,
        )
        self._attr_names += [
            "values",
            "descriptions",
        ]
        self._event_names += ["valuesChanged"]


class FilterOptionsWidget(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "filter-options-widget",
            **kwargs,
        )
        self._attr_names += [
            "modelValue",
            "options",
        ]
        self._event_names += [
            "update:modelValue",
        ]


class FilterOperatorWidget(HtmlElement):
    def __init__(self, **kwargs):
        super().__init__(
            "filter-operator-widget",
            **kwargs,
        )
        self._attr_names += [
            "operator",
            "invert",
        ]
        self._event_names += [
            "update:operator",
            "update:invert",
        ]



from pathlib import Path
import torchvision.transforms as transforms

BASE_PATH = str(Path(__file__).parent.absolute())

with open(Path(BASE_PATH, "imagenet_classes.txt")) as f:
    imagenet_categories = f.read().splitlines()

# Pytorch pre-processing
imagenet_model_loader = transforms.Compose(
    [
        transforms.ToPILImage(),
        transforms.Resize((224, 224)),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ]
)



from typing import (
    Any,
    Dict,
    List,
    Protocol,
    Tuple,
    Optional,
    Generic,
    TypeVar,
)
import numpy as np
from scipy.special import softmax
from sklearn.metrics.pairwise import cosine_similarity
from ubelt import grabdata

# pytorch
import torch
import torch.nn as nn
import torchvision.models as models

# xaitk-saliency
from smqtk_detection.impls.detect_image_objects.resnet_frcnn import ResNetFRCNN
from smqtk_detection.impls.detect_image_objects.centernet import CenterNetVisdrone

# labels + transform
from .assets import (
    imagenet_categories,
    imagenet_model_loader,
)

import logging
import warnings

logger = logging.getLogger("xaitks_saliency_demo")

# ResNetFRCNN produces some deprecation warnings
warnings.filterwarnings("ignore", category=UserWarning)

# -----------------------------------------------------------------------------

DEVICE = torch.device("cpu")


def update_ml_device(cpu_only=True):
    global DEVICE
    if not cpu_only and torch.cuda.is_available():
        DEVICE = torch.device("cuda")
        logger.info(" ~~~ Using GPU ~~~ \n")
    else:
        logger.info(" ~~~ Using CPU ~~~ \n")


# -----------------------------------------------------------------------------

__all__ = [
    # API ------------------
    "get_model",
    # classification -------
    "ClassificationResNet50",
    "ClassificationAlexNet",
    "ClassificationVgg16",
    # similarity -----------
    "SimilarityResNet50",
    "SimilarityAlexNet",
    "SimilarityVgg16",
    # detection ------------
    "DetectionFRCNN",
    "DetectionCenterNetVisdrone",
]

# -----------------------------------------------------------------------------
# Model file acquisition / caching
# -----------------------------------------------------------------------------

CENTERNET_RESNET50 = grabdata(
    "https://data.kitware.com/api/v1/item/623259f64acac99f426f21db/download",
    fname="centernet-resnet50.pth",
    appname="xaitk-saliency-demo",
    hash_prefix="a0083ec55d46c420d06c414e5ecc1863d6ad9b6a1732acff5c9dba28158a4"
    "c5a04f43541415d503fa776031a7329e3912864ae2348b3bee035df0d1e7a"
    "cefa49",
)

# -----------------------------------------------------------------------------
# Class normalizing model usage
# -----------------------------------------------------------------------------


class ModelProtocol(Protocol):
    @property
    def device(self) -> Any: ...

    @property
    def state(self) -> Any: ...

    @property
    def model(self) -> Any: ...


T = TypeVar("T", covariant=True)


class PredictProtocol(Protocol, Generic[T]):
    def predict(self, input) -> T: ...


class RunProtocol(Protocol):
    def run(self, input, *args) -> Dict[str, Any]: ...


class AbstractModel(ModelProtocol):
    def __init__(self, server, model, device=None):
        if device is None:
            device = DEVICE
        self._server = server
        self._device = device
        self.topk = 10
        self._model = model
        if isinstance(model, torch.nn.Module):
            self._model = model.to(self._device)
            self._model.eval()

    def __call__(self, *args, **kwargs):
        return self._model(*args, **kwargs)

    @property
    def device(self):
        return self._device

    @property
    def state(self):
        return self._server.state

    @property
    def model(self):
        return self._model


# -----------------------------------------------------------------------------
# Classification
# -----------------------------------------------------------------------------


class ResNetPredict(ModelProtocol, PredictProtocol, Protocol):
    @torch.no_grad()
    def predict(self, input) -> np.ndarray:
        """
        Run prediction with the set model and return the final output layer
        as a vector for one image.
        """
        input = imagenet_model_loader(input).unsqueeze(0)
        input = input.to(self.device)
        output = self.model(input).squeeze()
        return output.cpu().numpy()


class ClassificationRun(ModelProtocol, PredictProtocol, RunProtocol, Protocol):
    topk: np.ndarray

    def run(self, input, *_):
        preds = self.predict(input)
        preds = softmax(preds)
        topk = np.argsort(-preds)[: self.state.TOP_K]
        output = [(imagenet_categories[i], preds[i]) for i in topk]

        # store for later
        self.topk = topk

        logger.info(f"Predicted classes: {output}")
        return {
            "type": "classification",
            "topk": topk,
            "classes": output,
        }


class ClassificationResNet50(AbstractModel, ResNetPredict, ClassificationRun):
    def __init__(self, server):
        super().__init__(server, models.resnet50(pretrained=True))


class ClassificationAlexNet(AbstractModel, ResNetPredict, ClassificationRun):
    def __init__(self, server):
        super().__init__(server, models.alexnet(pretrained=True))


class ClassificationVgg16(AbstractModel, ResNetPredict, ClassificationRun):
    def __init__(self, server):
        super().__init__(server, models.vgg16(pretrained=True))


# -----------------------------------------------------------------------------
# Similarity
# -----------------------------------------------------------------------------


class SimilarityRun(ModelProtocol, PredictProtocol, RunProtocol, Protocol):
    def run(self, query, reference, *_):
        p_query = self.predict(query)
        p_reference = self.predict(reference)
        similarity = cosine_similarity(
            p_query.reshape(1, -1),
            p_reference.reshape(1, -1),
        ).item()
        logger.info(f"Similarity score: {similarity}")
        return {
            "type": "similarity",
            "similarity": similarity,
        }


class SimilarityResNet50(AbstractModel, ResNetPredict, SimilarityRun):
    def __init__(self, server):
        super().__init__(server, models.resnet50(pretrained=True))
        # Perform model surgery to get feature descriptors
        self._model = nn.Sequential(*list(self._model.children())[:-1])


class SimilarityAlexNet(AbstractModel, ResNetPredict, SimilarityRun):
    def __init__(self, server):
        super().__init__(server, models.alexnet(pretrained=True))
        # Perform model surgery to get feature descriptors
        self._model.classifier = nn.Sequential(*list(self._model.classifier.children())[:-1])


class SimilarityVgg16(AbstractModel, ResNetPredict, SimilarityRun):
    def __init__(self, server):
        super().__init__(server, models.vgg16(pretrained=True))
        # Perform model surgery to get feature descriptors
        self._model.classifier = nn.Sequential(*list(self._model.classifier.children())[:-1])


# -----------------------------------------------------------------------------
# Detection
# -----------------------------------------------------------------------------


class DetectionPredict(ModelProtocol, PredictProtocol, RunProtocol, Protocol):
    def predict(self, input: np.ndarray) -> Tuple[np.ndarray, np.ndarray, Optional[List[str]]]:
        """
        Predict detections for a single image, returning bounding-boxes, scores
        and labels as arrays (in that order).
        Bounding boxes are in (minX, minY, maxX, maxY) format.
        Boxes array is of shape (nDets x 4).
        Scores array is of shape (nDets x nClasses).
        Labels list is of length (nClasses).
        """
        preds = list(list(self.model([input]))[0])
        n_preds = len(preds)
        # TODO: What should happen when there are no detections in the image?
        n_classes = len(preds[0][1])
        # scores matrix should be ordered the same as `list(dict.keys())`
        bboxes = np.empty((n_preds, 4), dtype=np.float32)
        scores = np.empty((n_preds, n_classes), dtype=np.float32)
        labels = None
        for i, (bbox, score_dict) in enumerate(preds):
            bboxes[i] = (*bbox.min_vertex, *bbox.max_vertex)
            scores[i] = list(score_dict.values())
            if labels is None:
                labels = list(score_dict.keys())

        return bboxes, scores, labels


class DetectionRun(ModelProtocol, PredictProtocol, RunProtocol, Protocol):
    topk: np.ndarray

    def run(self, input: np.ndarray, *_):
        """
        Generic "run" function for producing top-K image object detections.
        """
        boxes, scores, labels = self.predict(input)
        max_scores = np.max(scores, axis=1)
        max_scores_idx = np.argmax(scores, axis=1)
        topk = np.argsort(-max_scores)[: self.state.TOP_K]
        output = [(labels[max_scores_idx[i]], max_scores[i], boxes[i]) for i in topk]
        # output is of format
        # [ ..., (category-label, score, LTRB bbox array), ... ]

        # store for later
        self.topk = topk

        logger.info(f"Predicted bounding boxes: {output}")
        return {
            "type": "detection",
            "topk": topk,
            "detection": output,
        }


class DetectionFRCNN(AbstractModel, DetectionPredict, DetectionRun):
    def __init__(self, server):
        d = DEVICE
        model = ResNetFRCNN(
            use_cuda=True if "cuda" in d.type.lower() else False,
            cuda_device=d.type + (f":{d.index}" if d.index is not None else ""),
        )
        super().__init__(server, model)


class DetectionCenterNetVisdrone(AbstractModel, DetectionPredict, DetectionRun):
    def __init__(self, server):
        d = DEVICE
        model = CenterNetVisdrone(
            arch="resnet50",
            model_file=CENTERNET_RESNET50,
            max_dets=500,
            use_cuda=True if "cuda" in d.type.lower() else False,
        )
        super().__init__(server, model)


# -----------------------------------------------------------------------------
# Factory instance maps
# -----------------------------------------------------------------------------

MODEL_INSTANCES: Dict[str, Dict[str, AbstractModel]] = {}

# -----------------------------------------------------------------------------
# Factory methods
# -----------------------------------------------------------------------------


def get_model(server, model_name):
    global MODEL_INSTANCES
    MODEL_INSTANCES.setdefault(server, {})

    if model_name in MODEL_INSTANCES[server]:
        return MODEL_INSTANCES[server][model_name]

    try:
        constructor = globals()[model_name]
        MODEL_INSTANCES[server][model_name] = constructor(server)
    except KeyError:
        logger.info(f"Could not find {model_name} in {globals().keys()}")

    return MODEL_INSTANCES[server].get(model_name)



from typing import Protocol, Iterable, Literal, Callable, FrozenSet, TypeVar, Generic

LogicalOperator = Literal["and", "or"]
OperatorFunction = Callable[[bool, bool], bool]


def get_operator_function(operator: LogicalOperator) -> OperatorFunction:
    if operator == "and":
        return lambda a, b: a and b
    else:
        return lambda a, b: a or b


T = TypeVar("T", contravariant=True)


class FilterProtocol(Protocol, Generic[T]):
    def evaluate(self, item: T) -> bool: ...


class NoneFilter(FilterProtocol, Generic[T]):
    def evaluate(self, item: T) -> bool:
        return True


class NotFilter(FilterProtocol, Generic[T]):
    def __init__(self, filter: FilterProtocol):
        self.filter = filter

    def evaluate(self, item: T) -> bool:
        return not self.filter.evaluate(item)


class ComposableFilter(FilterProtocol[T]):
    filter_a: FilterProtocol[T]
    filter_b: FilterProtocol[T]
    operator: LogicalOperator
    operator_fn: OperatorFunction

    def __init__(self):
        self.compose(NoneFilter(), NoneFilter(), "and")

    def compose(self, a: FilterProtocol[T], b: FilterProtocol[T], operator: LogicalOperator):
        self.filter_a = a
        self.filter_b = b
        self.operator = operator
        self.operator_fn = get_operator_function(self.operator)

    def evaluate(self, item: T) -> bool:
        return self.operator_fn(self.filter_a.evaluate(item), self.filter_b.evaluate(item))


class ConcreteIdFilter(FilterProtocol[Iterable[int]]):
    ids: FrozenSet[int]
    operator: LogicalOperator
    operator_fn: OperatorFunction

    def __init__(self):
        self.set_ids([], "and")

    def set_ids(self, ids: Iterable[int], operator: LogicalOperator):
        self.ids = frozenset(ids)
        self.operator = operator

    def evaluate(self, item: Iterable[int]) -> bool:
        if len(self.ids) == 0:
            return True

        matches = 0

        unique_item_ids = frozenset(item)

        for id in unique_item_ids:
            if id in self.ids:
                matches += 1

        if self.operator == "and":
            return matches == len(self.ids)
        else:
            return matches > 0



from nrtk_explorer.library import object_detector

import json
import os
import nrtk_explorer.test_data

DIR_PATH = os.path.dirname(nrtk_explorer.test_data.__file__)
DATASET_PATH = f"{DIR_PATH}/coco-od-2017/"
DATASET = f"{DATASET_PATH}/test_val2017.json"


def test_detector_small():
    ds = json.load(open(DATASET))
    sample = [f"{DATASET_PATH}/{img['file_name']}" for img in ds["images"]][:15]
    detector = object_detector.ObjectDetector(model_name="hustvl/yolos-tiny")
    img = detector.eval(paths=sample)
    assert len(img) == 15



from .core import create_engine


def jupyter_proxy_info():
    """Get the config to run the trame application via jupyter's server proxy

    This is provided to the `jupyter_serverproxy_servers` entrypoint, and the
    jupyter server proxy will use it to start the application as a separate
    process.
    """
    return {
        "command": ["nrtk_explorer", "-p", "0", "--server"],
    }


def show(server=None, **kwargs):
    """Show application into a Jupyter cell"""
    app = create_engine(server)
    app.show_in_jupyter(**kwargs)



from trame_server.core import Server, State, Controller


class Applet:
    def __init__(self, server: Server):
        self._server = server

    @property
    def server(self) -> Server:
        return self._server

    @property
    def state(self) -> State:
        return self.server.state

    @property
    def ctrl(self) -> Controller:
        return self.server.controller

    @property
    def context(self) -> State:
        return self.server.context



r"""
Define your classes and create the instances that you need to expose
"""

import logging
from typing import Iterable
from trame.app import get_server
from trame.ui.quasar import QLayout
from trame.widgets import quasar
from trame.widgets import html
from trame_server.utils.namespace import Translator
from nrtk_explorer.library import images_manager
from nrtk_explorer.library.filtering import FilterProtocol

from nrtk_explorer.app.embeddings import EmbeddingsApp
from nrtk_explorer.app.transforms import TransformsApp
from nrtk_explorer.app.filtering import FilteringApp
from nrtk_explorer.app.applet import Applet
from nrtk_explorer.app.ui.collapsible_card import collapsible_card
import nrtk_explorer.test_data
from pathlib import Path

import os

import json
import random


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

html.Template.slot_names.add("before")
html.Template.slot_names.add("after")

HORIZONTAL_SPLIT_DEFAULT_VALUE = 17
VERTICAL_SPLIT_DEFAULT_VALUE = 40

DIR_NAME = os.path.dirname(nrtk_explorer.test_data.__file__)
DEFAULT_DATASETS = [
    f"{DIR_NAME}/coco-od-2017/test_val2017.json",
    f"{DIR_NAME}/OIRDS_v1_0/oirds.json",
]


def image_id_to_meta(image_id):
    return f"{image_id}_meta"


def image_id_to_result(image_id):
    return f"{image_id}_result"


def parse_dataset_dirs(datasets):
    return [{"label": Path(ds).name, "value": ds} for ds in datasets]


# ---------------------------------------------------------
# Engine class
# ---------------------------------------------------------


class Engine(Applet):
    def __init__(self, server=None):
        if server is None:
            server = get_server()

        super().__init__(server)

        self.server.cli.add_argument(
            "--dataset",
            nargs="+",
            default=DEFAULT_DATASETS,
            help="Path of the json file describing the image dataset",
        )

        known_args, _ = self.server.cli.parse_known_args()
        self.input_paths = known_args.dataset
        self.state.current_dataset = self.input_paths[0]

        self.context["image_objects"] = {}
        self.context["images_manager"] = images_manager.ImagesManager()
        self.context["annotations"] = {}

        self._ui = None

        self.state.collapse_dataset = False
        self.state.collapse_embeddings = False
        self.state.collapse_filter = False
        self.state.collapse_transforms = False
        self.state.client_only(
            "collapse_dataset", "collapse_embeddings", "collapse_filter", "collapse_transforms"
        )

        self.state.horizontal_split = HORIZONTAL_SPLIT_DEFAULT_VALUE
        self.state.vertical_split = VERTICAL_SPLIT_DEFAULT_VALUE
        self.state.client_only("horizontal_split", "vertical_split")

        transforms_translator = Translator()
        transforms_translator.add_translation(
            "feature_extraction_model", "current_transforms_model"
        )

        self._transforms_app = TransformsApp(
            server=self.server.create_child_server(translator=transforms_translator)
        )

        embeddings_translator = Translator()
        embeddings_translator.add_translation(
            "feature_extraction_model", "current_embeddings_model"
        )

        self._embeddings_app = EmbeddingsApp(
            server=self.server.create_child_server(translator=embeddings_translator),
        )

        filtering_translator = Translator()
        filtering_translator.add_translation("categories", "annotation_categories")
        self._filtering_app = FilteringApp(
            server=self.server.create_child_server(translator=filtering_translator),
        )

        self._embeddings_app.set_on_select(self._transforms_app.on_selected_images_change)
        self._transforms_app.set_on_transform(self._embeddings_app.on_run_transformations)
        self._embeddings_app.set_on_hover(self._transforms_app.on_image_hovered)
        self._transforms_app.set_on_hover(self._embeddings_app.on_image_hovered)
        self._filtering_app.set_on_apply_filter(self.on_filter_apply)

        # Set state variable
        self.state.trame__title = "nrtk_explorer"

        # Bind instance methods to controller
        self.ctrl.on_server_reload = self.ui
        self.ctrl.add("on_server_ready")(self.on_server_ready)

        self.state.num_images_max = 0
        self.state.num_images_disabled = True
        self.state.random_sampling = False
        self.state.random_sampling_disabled = True
        self.state.images_id = []

        # Generate UI
        self.ui()
        self.context.images_manager = images_manager.ImagesManager()

    def on_server_ready(self, *args, **kwargs):
        # Bind instance methods to state change
        self.state.change("current_dataset")(self.on_dataset_change)
        self.state.change("num_images")(self.on_num_images_change)
        self.state.change("random_sampling")(self.on_random_sampling_change)

        self.on_dataset_change()

    def on_dataset_change(self, **kwargs):
        # Reset cache
        self.context.images_manager = images_manager.ImagesManager()

        with open(self.state.current_dataset) as f:
            dataset = json.load(f)

        self.state.num_images_max = len(dataset["images"])
        self.state.random_sampling_disabled = False
        self.state.num_images_disabled = False

        self.reload_images()

    def on_filter_apply(self, filter: FilterProtocol[Iterable[int]], **kwargs):
        selected_indices = []

        for index, image_id in enumerate(self.state.images_ids):
            image_annotations_categories = map(
                lambda annotation: annotation["category_id"],
                self.context["annotations"].get(f"img_{image_id}", []),
            )

            include = filter.evaluate(image_annotations_categories)

            if include:
                selected_indices.append(index)

        self._embeddings_app.on_select(selected_indices)

    def on_num_images_change(self, **kwargs):
        self.reload_images()

    def on_random_sampling_change(self, **kwargs):
        self.reload_images()

    def reload_images(self):
        with open(self.state.current_dataset) as f:
            dataset = json.load(f)

        categories = {}
        for category in dataset["categories"]:
            categories[category["id"]] = category

        images = dataset["images"]

        selected_images = []
        if self.state.num_images:
            if self.state.random_sampling:
                selected_images = random.sample(images, self.state.num_images)
            else:
                selected_images = images[: self.state.num_images]
        else:
            selected_images = images

        paths = list()
        for image in selected_images:
            paths.append(
                os.path.join(
                    os.path.dirname(self.state.current_dataset),
                    image["file_name"],
                )
            )

        self.context.paths = paths
        self.state.annotation_categories = categories
        self.state.images_ids = [img["id"] for img in selected_images]

    def ui(self, *args, **kwargs):
        if self._ui is None:
            with QLayout(
                self.server, view="lhh LpR lff", classes="shadow-2 rounded-borders bg-grey-2"
            ) as layout:
                # # Toolbar
                with quasar.QHeader():
                    with quasar.QToolbar(classes="shadow-4"):
                        quasar.QToolbarTitle("NRTK_EXPLORER")

                # # Main content
                with quasar.QPageContainer():
                    with quasar.QPage():
                        with quasar.QSplitter(
                            model_value=("horizontal_split",),
                            classes="inherit-height",
                            before_class="inherit-height zero-height scroll",
                            after_class="inherit-height zero-height",
                        ):
                            with html.Template(v_slot_before=True):
                                with html.Div(classes="q-pa-md q-gutter-md"):
                                    (
                                        dataset_title_slot,
                                        dataset_content_slot,
                                        dataset_actions_slot,
                                    ) = collapsible_card("collapse_dataset")

                                    with dataset_title_slot:
                                        html.Span("Dataset Selection", classes="text-h6")

                                    with dataset_content_slot:
                                        quasar.QSelect(
                                            label="Dataset",
                                            v_model=("current_dataset",),
                                            options=(parse_dataset_dirs(self.input_paths),),
                                            filled=True,
                                            emit_value=True,
                                            map_options=True,
                                            dense=True,
                                        )
                                        quasar.QSlider(
                                            v_model=("num_images", 15),
                                            min=(0,),
                                            max=("num_images_max", 25),
                                            disable=("num_images_disabled", True),
                                            step=(1,),
                                        )
                                        html.P(
                                            "{{num_images}}/{{num_images_max}} images",
                                            classes="text-caption text-center",
                                        )

                                        quasar.QToggle(
                                            v_model=("random_sampling", False),
                                            dense=False,
                                            label="Random selection",
                                        )

                                    (
                                        embeddings_title_slot,
                                        embeddings_content_slot,
                                        embeddings_actions_slot,
                                    ) = collapsible_card("collapse_embeddings")

                                    with embeddings_title_slot:
                                        html.Span("Embeddings", classes="text-h6")

                                    with embeddings_content_slot:
                                        self._embeddings_app.settings_widget()

                                    with embeddings_actions_slot:
                                        self._embeddings_app.compute_ui()

                                    filter_title_slot, filter_content_slot, filter_actions_slot = (
                                        collapsible_card("collapse_filter")
                                    )

                                    with filter_title_slot:
                                        html.Span("Category Filter", classes="text-h6")

                                    with filter_content_slot:
                                        self._filtering_app.filter_operator_ui()
                                        self._filtering_app.filter_options_ui()

                                    with filter_actions_slot:
                                        self._filtering_app.filter_apply_ui()

                                    (
                                        transforms_title_slot,
                                        transforms_content_slot,
                                        transforms_actions_slot,
                                    ) = collapsible_card("collapse_transforms")

                                    with transforms_title_slot:
                                        html.Span("Transform Settings", classes="text-h6")

                                    with transforms_content_slot:
                                        self._transforms_app.settings_widget()

                                    with transforms_actions_slot:
                                        self._transforms_app.apply_ui()

                            with html.Template(v_slot_after=True):
                                with quasar.QSplitter(
                                    v_model=("vertical_split",),
                                    limits=("[0,100]",),
                                    horizontal=True,
                                    classes="inherit-height zero-height",
                                    before_class="q-pa-md",
                                    after_class="q-pa-md",
                                ):
                                    with html.Template(v_slot_before=True):
                                        self._embeddings_app.visualization_widget()

                                    with html.Template(v_slot_after=True):
                                        with html.Div(classes="row q-col-gutter-md"):
                                            with html.Div(classes="col-6"):
                                                with quasar.QCard(flat=True, bordered=True):
                                                    with quasar.QCardSection():
                                                        html.Span(
                                                            "Original Dataset", classes="text-h5"
                                                        )

                                                    with quasar.QCardSection():
                                                        self._transforms_app.original_dataset_widget()

                                            with html.Div(classes="col-6"):
                                                with quasar.QCard(
                                                    flat=True,
                                                    bordered=True,
                                                    style="background-color: #ffcdd2;",
                                                ):
                                                    with quasar.QCardSection():
                                                        html.Span(
                                                            "Transformed Dataset",
                                                            classes="text-h5",
                                                        )

                                                    with quasar.QCardSection():
                                                        self._transforms_app.transformed_dataset_widget()

            self._ui = layout

        return self._ui


def create_engine(server=None):
    # Get or create server
    if server is None:
        server = get_server()

    if isinstance(server, str):
        server = get_server(server)

    server.client_type = "vue3"

    return Engine(server)



import logging
import operator
import torch
import transformers

from functools import reduce
from typing import Optional

from nrtk_explorer.library import images_manager

Annotations = list[list[dict]]


class ObjectDetector:
    """Object detection using Hugging Face's transformers library"""

    def __init__(
        self,
        model_name: str = "facebook/detr-resnet-50",
        manager: Optional[images_manager.ImagesManager] = None,
        force_cpu: bool = False,
    ):
        if manager is None:
            manager = images_manager.ImagesManager()

        self.manager = manager
        self.device = "cuda" if torch.cuda.is_available() and not force_cpu else "cpu"
        self.pipeline = model_name

    @property
    def device(self) -> str:
        """Get the device to use for feature extraction"""
        return str(self._device)

    @device.setter
    def device(self, dev: str):
        """Set the device to use for feature extraction"""
        logging.info(f"Using {dev} devices for feature extraction")
        self._device = torch.device(dev)

    @property
    def pipeline(self) -> transformers.pipeline:
        """Get the pipeline for object detection using Hugging Face's transformers library"""
        return self._pipeline

    @pipeline.setter
    def pipeline(self, model_name: str):
        """Set the pipeline for object detection using Hugging Face's transformers library"""
        self._pipeline = transformers.pipeline(model=model_name, device=self.device)
        # Do not display warnings
        transformers.utils.logging.set_verbosity_error()

    def eval(
        self,
        paths: list[str],
        content: Optional[dict] = None,
        batch_size: int = 32,
    ) -> Annotations:
        """Compute object recognition, return it in a dictionary of COCO format"""
        if len(paths) == 0:
            return []

        images: dict = {}

        # Group images by size (shape)
        for path in paths:
            img = None
            if content and path in content:
                img = content[path]
            else:
                img = self.manager.load_image(path)

            images.setdefault(img.size, []).append(img)

        # Call by each group
        predictions = [self.pipeline(group, batch_size=batch_size) for group in images.values()]
        # Flatten the list of predictions
        return reduce(operator.iadd, predictions)



r"""
Define your classes and create the instances that you need to expose
"""

import logging
from typing import Dict

from trame.ui.quasar import QLayout
from trame.widgets import quasar
from trame.widgets import html
from trame.app import get_server

import nrtk_explorer.library.transforms as trans
import nrtk_explorer.library.nrtk_transforms as nrtk_trans
from nrtk_explorer.library import images_manager, object_detector
from nrtk_explorer.app.ui.image_list import image_list_component
from nrtk_explorer.app.applet import Applet
from nrtk_explorer.app.parameters import ParametersApp
from nrtk_explorer.library.ml_models import (
    ClassificationResNet50,
    ClassificationAlexNet,
    ClassificationVgg16,
)
import nrtk_explorer.test_data

import json
import os


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def image_id_to_meta(image_id):
    return f"{image_id}_meta"


def image_id_to_result(image_id):
    return f"{image_id}_result"


# ---------------------------------------------------------
# Engine class
# ---------------------------------------------------------


DIR_NAME = os.path.dirname(nrtk_explorer.test_data.__file__)
DATASET_DIRS = [
    f"{DIR_NAME}/OIRDS_v1_0/oirds.json",
    f"{DIR_NAME}/OIRDS_v1_0/oirds_test.json",
    f"{DIR_NAME}/OIRDS_v1_0/oirds_train.json",
]


class TransformsApp(Applet):
    def __init__(self, server):
        super().__init__(server)

        self._parameters_app = ParametersApp(
            server=server,
        )

        self._parameters_app.on_apply_transform = self.on_apply_transform

        self._ui = None

        self.is_standalone_app = self.server.state.parent is None
        if self.is_standalone_app:
            self.context.images_manager = images_manager.ImagesManager()

        if self.context["image_objects"] is None:
            self.context["image_objects"] = {}

        self.models = {
            "ClassificationResNet50": ClassificationResNet50(server),
            "ClassificationAlexNet": ClassificationAlexNet(server),
            "ClassificationVgg16": ClassificationVgg16(server),
        }

        self._on_transform_fn = None
        self.state.models = [k for k in self.models.keys()]
        self.state.feature_extraction_model = self.state.models[0]

        self._transforms: Dict[str, trans.ImageTransform] = {
            "identity": trans.IdentityTransform(),
            "blur": trans.GaussianBlurTransform(),
            "invert": trans.InvertTransform(),
            "downsample": trans.DownSampleTransform(),
            "nrtk_blur": nrtk_trans.NrtkGaussianBlurTransform(),
            "nrtk_pybsm": nrtk_trans.NrtkPybsmTransform(),
        }

        self._parameters_app._transforms = self._transforms

        self.state.annotation_categories = {}

        self.state.source_image_ids = []
        self.state.transformed_image_ids = []
        self.state.transforms = [k for k in self._transforms.keys()]
        self.state.current_transform = self.state.transforms[0]

        if self.state.current_dataset is None:
            self.state.current_dataset = DATASET_DIRS[0]

        self.state.current_num_elements = 15

        self.server.controller.add("on_server_ready")(self.on_server_ready)
        self._on_hover_fn = None
        self.detector = object_detector.ObjectDetector(model_name="hustvl/yolos-tiny")

    def on_server_ready(self, *args, **kwargs):
        # Bind instance methods to state change
        self.state.change("feature_extraction_model")(self.on_feature_extraction_model_change)
        self.state.change("current_dataset")(self.on_current_dataset_change)
        self.state.change("current_num_elements")(self.on_current_num_elements_change)

        self.on_current_dataset_change(self.state.current_dataset)

    def set_on_transform(self, fn):
        self._on_transform_fn = fn

    def on_transform(self, *args, **kwargs):
        if self._on_transform_fn:
            self._on_transform_fn(*args, **kwargs)

    def on_apply_transform(self, *args, **kwargs):
        logger.info(">>> ENGINE(a): on_apply_transform")

        current_transform = self.state.current_transform
        transformed_image_ids = []
        transform = self._transforms[current_transform]
        for image_id in self.state.source_image_ids:
            image = self.context["image_objects"][image_id]

            transformed_image_id = f"transformed_{image_id}"
            meta_id = image_id_to_meta(image_id)
            transformed_meta_id = image_id_to_meta(transformed_image_id)

            transformed_img = transform.execute(image)

            self.context["image_objects"][transformed_image_id] = transformed_img

            transformed_image_ids.append(transformed_image_id)

            self.state[transformed_image_id] = images_manager.convert_to_base64(transformed_img)
            self.state[transformed_meta_id] = self.state[meta_id]

            self.state.hovered_id = -1

        self.state.transformed_image_ids = transformed_image_ids
        self.compute_annotations(transformed_image_ids)

        # Only invoke callbacks when we transform images
        if len(transformed_image_ids) > 0:
            self.on_transform(transformed_image_ids)

    def compute_annotations(self, ids):
        """Compute annotations for the given image ids using the object detector model."""
        if len(ids) == 0:
            return

        for id_ in ids:
            self.context["annotations"][id_] = []

        prediction = self.detector.eval(paths=ids, content=self.context.image_objects)

        for id_, annotations in zip(ids, prediction):
            image_annotations = self.context["annotations"].setdefault(id_, [])
            for prediction in annotations:
                category_id = 0
                for cat_id, cat in self.state.annotation_categories.items():
                    if cat["name"] == prediction["label"]:
                        category_id = cat_id

                bbox = prediction["box"]
                image_annotations.append(
                    {
                        "category_id": category_id,
                        "id": category_id,
                        "bbox": [
                            bbox["xmin"],
                            bbox["ymin"],
                            bbox["xmax"] - bbox["xmin"],
                            bbox["ymax"] - bbox["ymin"],
                        ],
                    }
                )

        self.update_model_result(ids, self.state.feature_extraction_model)

    def on_current_num_elements_change(self, current_num_elements, **kwargs):
        with open(self.state.current_dataset) as f:
            dataset = json.load(f)
        ids = [img["id"] for img in dataset["images"]]
        return self.on_selected_images_change(ids[:current_num_elements])

    def on_selected_images_change(self, selected_ids):
        source_image_ids = []

        current_dir = os.path.dirname(self.state.current_dataset)

        with open(self.state.current_dataset) as f:
            dataset = json.load(f)

        for selected_id in selected_ids:
            image_index = self.context.image_id_to_index[selected_id]
            if image_index >= len(dataset["images"]):
                continue

            image_metadata = dataset["images"][image_index]

            image_id = f"img_{image_metadata['id']}"
            meta_id = image_id_to_meta(image_id)

            source_image_ids.append(image_id)

            image_filename = os.path.join(current_dir, image_metadata["file_name"])

            img = self.context.images_manager.load_image(image_filename)

            self.state[image_id] = images_manager.convert_to_base64(img)
            self.state[meta_id] = {
                "width": image_metadata["width"],
                "height": image_metadata["height"],
            }
            self.state.hovered_id = -1

            self.context.image_objects[image_id] = img

        self.state.source_image_ids = source_image_ids
        self.compute_annotations(source_image_ids)
        self.update_model_result(self.state.source_image_ids, self.state.feature_extraction_model)
        self.on_apply_transform()

    def reset_data(self):
        source_image_ids = self.state.source_image_ids
        transformed_image_ids = self.state.transformed_image_ids

        self.state.source_image_ids = []
        self.state.transformed_image_ids = []
        self.state.annotation_categories = {}

        for image_id in source_image_ids:
            result_id = image_id_to_result(image_id)
            meta_id = image_id_to_meta(image_id)

            if self.state.has(image_id) and self.state[image_id] is not None:
                self.state[image_id] = None

            if self.state.has(result_id) and self.state[result_id] is not None:
                self.state[result_id] = None

            if self.state.has(meta_id) and self.state[meta_id] is not None:
                self.state[meta_id] = None

            if image_id in self.context["image_objects"]:
                del self.context["image_objects"][image_id]

        for image_id in transformed_image_ids:
            result_id = image_id_to_result(image_id)
            meta_id = image_id_to_meta(image_id)

            if self.state.has(image_id) and self.state[image_id] is not None:
                self.state[image_id] = None

            if self.state.has(result_id) and self.state[result_id] is not None:
                self.state[result_id] = None

            if self.state.has(meta_id) and self.state[meta_id] is not None:
                self.state[meta_id] = None

            if image_id in self.context["image_objects"]:
                del self.context["image_objects"][image_id]

    def on_current_dataset_change(self, current_dataset, **kwargs):
        logger.info(f">>> ENGINE(a): on_current_dataset_change change {self.state}")

        self.reset_data()

        with open(current_dataset) as f:
            dataset = json.load(f)

        categories = {}

        for category in dataset["categories"]:
            categories[category["id"]] = category

        self.state.annotation_categories = categories

        self.context["annotations"] = {}

        self.context.image_id_to_index = {}
        for i, image in enumerate(dataset["images"]):
            self.context.image_id_to_index[image["id"]] = i

        if self.is_standalone_app:
            self.context.images_manager = images_manager.ImagesManager()

    def on_feature_extraction_model_change(self, **kwargs):
        logger.info(f">>> ENGINE(a): on_feature_extraction_model_change change {self.state}")

        feature_extraction_model = self.state.feature_extraction_model

        self.update_model_result(self.state.source_image_ids, feature_extraction_model)
        self.update_model_result(self.state.transformed_image_ids, feature_extraction_model)

    def update_model_result(self, image_ids, feature_extraction_model):
        for image_id in image_ids:
            result_id = image_id_to_result(image_id)
            self.state[result_id] = self.context["annotations"].get(image_id, [])

    def on_image_hovered(self, index):
        self.state.hovered_id = index

    def set_on_hover(self, fn):
        self._on_hover_fn = fn

    def on_hover(self, hover_event):
        id_ = int(hover_event["id"])
        is_transformation = bool(hover_event["isTransformation"])
        self.on_image_hovered(id_)
        if self._on_hover_fn:
            self._on_hover_fn(id_, is_transformation)

    def settings_widget(self):
        with html.Div(trame_server=self.server):
            with html.Div(classes="col"):
                quasar.QSelect(
                    label="Object detection Model",
                    v_model=("object_detection_model", "facebook/detr-resnet-50"),
                    options=(
                        [
                            {
                                "label": "facebook/detr-resnet-50",
                                "value": "facebook/detr-resnet-50",
                            },
                        ],
                    ),
                    filled=True,
                    emit_value=True,
                    map_options=True,
                )

                self._parameters_app.transform_select_ui()

                with html.Div(
                    classes="q-pa-md q-ma-md",
                    style="border-style: solid; border-width: thin; border-radius: 0.5rem; border-color: lightgray;",
                ):
                    self._parameters_app.transform_params_ui()

    def apply_ui(self):
        with html.Div(trame_server=self.server):
            self._parameters_app.transform_apply_ui()

    def original_dataset_widget(self):
        with html.Div(trame_server=self.server):
            image_list_component("source_image_ids", self.on_hover)

    def transformed_dataset_widget(self):
        with html.Div(trame_server=self.server):
            image_list_component("transformed_image_ids", self.on_hover, is_transformation=True)

    # This is only used within when this module (file) is executed as an Standalone app.
    @property
    def ui(self):
        if self._ui is None:
            with QLayout(
                self.server, view="lhh LpR lff", classes="shadow-2 rounded-borders bg-grey-2"
            ) as layout:
                # # Toolbar
                with quasar.QHeader():
                    with quasar.QToolbar(classes="shadow-4"):
                        quasar.QBtn(
                            flat=True,
                            click="drawerLeft = !drawerLeft",
                            round=True,
                            dense=False,
                            icon="menu",
                        )
                        quasar.QToolbarTitle("Transforms")

                # # Main content
                with quasar.QPageContainer():
                    with quasar.QPage():
                        with html.Div(classes="row"):
                            with html.Div(classes="col-2 q-pa-md"):
                                with html.Div(
                                    classes="column justify-center", style="padding:1rem"
                                ):
                                    with html.Div(classes="col"):
                                        quasar.QSelect(
                                            label="Dataset",
                                            v_model=("current_dataset",),
                                            options=(DATASET_DIRS,),
                                            filled=True,
                                            emit_value=True,
                                            map_options=True,
                                        )

                                        html.P("Number of elements:", classes="text-body2")
                                        quasar.QSlider(
                                            v_model=("current_num_elements",),
                                            min=(0,),
                                            max=(25,),
                                            step=(1,),
                                            label=True,
                                            label_always=True,
                                        )
                                self.settings_widget()
                                self.apply_ui()

                            with html.Div(classes="col-5 q-pa-md"):
                                with html.Div(classes="row"):
                                    with html.Div(classes="col q-pa-md"):
                                        self.original_dataset_widget()

                            with html.Div(
                                classes="col-5 q-pa-md",
                                style="background-color: #ffcdcd;",
                            ):
                                with html.Div(classes="row"):
                                    with html.Div(classes="col q-pa-md"):
                                        self.transformed_dataset_widget()

                self._ui = layout
        return self._ui


def transforms(server=None, *args, **kwargs):
    server = get_server()
    server.client_type = "vue3"

    transforms_app = TransformsApp(server)
    transforms_app.ui

    server.start(**kwargs)


if __name__ == "__main__":
    transforms()



from typing import List
from nrtk_explorer.library import filtering

a, b, c = 0, 1, 2
test_values: List[List[int]] = [
    [],
    [a],
    [b],
    [c],
    [a, b],
    [a, c],
    [b, c],
    [a, b, c],
]


def test_concrete_and():
    filter = filtering.ConcreteIdFilter()
    result = list(map(filter.evaluate, test_values))
    assert result == [True] * 8

    filter.set_ids([], "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [True] * 8

    filter.set_ids((a,), "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, False, False, True, True, False, True]

    filter.set_ids((b,), "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, True, False, True, False, True, True]

    filter.set_ids((c,), "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, True, False, True, True, True]

    filter.set_ids(
        (
            a,
            b,
        ),
        "and",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, False, True, False, False, True]

    filter.set_ids(
        (
            a,
            c,
        ),
        "and",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, False, False, True, False, True]

    filter.set_ids(
        (
            b,
            c,
        ),
        "and",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, False, False, False, True, True]

    filter.set_ids(
        (
            a,
            b,
            c,
        ),
        "and",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, False, False, False, False, True]


def test_concrete_or():
    filter = filtering.ConcreteIdFilter()
    result = list(map(filter.evaluate, test_values))
    assert result == [True] * 8

    filter.set_ids([], "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [True] * 8

    filter.set_ids((a,), "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, False, False, True, True, False, True]

    filter.set_ids((b,), "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, True, False, True, False, True, True]

    filter.set_ids((c,), "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, True, False, True, True, True]

    filter.set_ids(
        (
            a,
            b,
        ),
        "or",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, True, False, True, True, True, True]

    filter.set_ids(
        (
            a,
            c,
        ),
        "or",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, False, True, True, True, True, True]

    filter.set_ids(
        (
            b,
            c,
        ),
        "or",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, True, True, True, True, True, True]

    filter.set_ids(
        (
            a,
            b,
            c,
        ),
        "or",
    )
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, True, True, True, True, True, True]


def test_concrete_not():
    filter = filtering.ConcreteIdFilter()
    filter.set_ids((a,), "and")
    not_filter = filtering.NotFilter(filter)
    result = list(map(not_filter.evaluate, test_values))
    assert result == [True, False, True, True, False, False, True, False]


def test_composite_and():
    filter_a = filtering.ConcreteIdFilter()
    filter_a.set_ids((a,), "and")

    filter_b = filtering.ConcreteIdFilter()
    filter_b.set_ids((b,), "and")

    filter_not_a = filtering.NotFilter(filter_a)
    filter_not_b = filtering.NotFilter(filter_b)

    filter = filtering.ComposableFilter()
    not_filter = filtering.NotFilter(filter)

    filter.compose(filter_a, filter_b, "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, False, False, False, True, False, False, True]
    result = list(map(not_filter.evaluate, test_values))
    assert result == [True, True, True, True, False, True, True, False]

    filter.compose(filter_a, filter_not_b, "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, False, False, False, True, False, False]

    filter.compose(filter_not_a, filter_not_b, "and")
    result = list(map(filter.evaluate, test_values))
    assert result == [True, False, False, True, False, False, False, False]


def test_composite_or():
    filter_a = filtering.ConcreteIdFilter()
    filter_a.set_ids((a,), "and")

    filter_b = filtering.ConcreteIdFilter()
    filter_b.set_ids((b,), "and")

    filter_not_a = filtering.NotFilter(filter_a)
    filter_not_b = filtering.NotFilter(filter_b)

    filter = filtering.ComposableFilter()

    filter.compose(filter_a, filter_b, "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [False, True, True, False, True, True, True, True]

    filter.compose(filter_a, filter_not_b, "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [True, True, False, True, True, True, False, True]

    filter.compose(filter_not_a, filter_not_b, "or")
    result = list(map(filter.evaluate, test_values))
    assert result == [True, True, True, True, False, True, True, False]



from PIL.Image import Image
from PIL import Image as ImageModule

import base64
import copy
import io

# Resolution for images to be used in model
IMAGE_MODEL_RESOLUTION = (224, 224)
THUMBNAIL_RESOLUTION = (250, 250)


def convert_to_base64(img: Image) -> str:
    """Convert image to base64 string"""
    buf = io.BytesIO()
    img.save(buf, format="png")
    return "data:image/png;base64," + base64.b64encode(buf.getvalue()).decode()


class ImagesManager:
    """Class to manage images and thumbnails"""

    def __init__(self):
        self.images = {}
        self.thumbnails = {}
        self.images_for_model = {}

    def prepare_for_model(self, img):
        """Prepare image for model input"""
        img = img.resize(IMAGE_MODEL_RESOLUTION)
        return img.convert("RGB")

    def load_image(self, path):
        """Load image from path and store it in cache if not already loaded"""
        if path not in self.images:
            self.images[path] = ImageModule.open(path)

        return self.images[path]

    def load_image_for_model(self, path):
        """Load image for model from path and store it in cache if not already loaded"""
        if path not in self.images_for_model:
            img = copy.copy(self.load_image(path))
            self.images_for_model[path] = self.prepare_for_model(img)

        return self.images_for_model[path]

    def load_thumbnail(self, path):
        """Load thumbnail from path and store it in cache if not already loaded"""
        if path not in self.thumbnails:
            img = copy.copy(self.load_image(path))
            img.thumbnail(THUMBNAIL_RESOLUTION)
            self.thumbnails[path] = img

        return self.thumbnails[path]



from typing import Callable, Iterable

from nrtk_explorer.app.applet import Applet

from trame.widgets import quasar, html
from trame.ui.quasar import QLayout

from trame.app import get_server

from nrtk_explorer.library.filtering import (
    FilterProtocol,
    NotFilter,
    ConcreteIdFilter,
)

from nrtk_explorer.widgets.nrtk_explorer import FilterOptionsWidget, FilterOperatorWidget


class FilteringApp(Applet):
    def __init__(self, server):
        super().__init__(server)

        self._on_apply_filter = lambda _filter: None

        self.state.categories = {
            0: {"id": 0, "name": "a"},
            1: {"id": 1, "name": "b"},
            2: {"id": 2, "name": "c"},
            3: {"id": 3, "name": "d"},
            4: {"id": 4, "name": "e"},
            5: {"id": 5, "name": "f"},
        }
        self.state.filter_categories = []
        self.state.filter_operator = "or"
        self.state.filter_not = False

        self.server.controller.add("on_server_ready")(self.on_server_ready)

        self._filter = ConcreteIdFilter()
        self._not_filter = NotFilter(self._filter)

        self._ui = None

    def set_on_apply_filter(self, fn: Callable[[FilterProtocol[Iterable[int]]], None]):
        self._on_apply_filter = fn

    def on_server_ready(self, *args, **kwargs):
        # Bind instance methods to state change
        self.on_filter_categories_change()
        self.state.change("filter_categories")(self.on_filter_categories_change)
        self.state.change("filter_operator")(self.on_filter_categories_change)

    def on_apply_click(self):
        if self.state.filter_not:
            self._on_apply_filter(self._not_filter)
        else:
            self._on_apply_filter(self._filter)

    def on_filter_categories_change(self, **kwargs):
        self._filter.set_ids(self.state.filter_categories, self.state.filter_operator)

    def on_update_operator(self, operator, **kwargs):
        self.state.filter_operator = operator

    def on_update_filter_not(self, filter_not):
        self.state.filter_not = filter_not

    def filter_options_ui(self):
        with html.Div(trame_server=self.server):
            FilterOptionsWidget(
                v_model=("filter_categories",),
                options=("categories",),
            )

    def filter_operator_ui(self):
        with html.Div(trame_server=self.server):
            FilterOperatorWidget(
                operator=("filter_operator",),
                invert=("filter_not",),
                **{
                    "update:operator": (self.on_update_operator, "[$event]"),
                    "update:invert": (self.on_update_filter_not, "[$event]"),
                },
            )

    def filter_apply_ui(self):
        with html.Div(trame_server=self.server):
            quasar.QBtn(
                "Apply",
                click=(self.on_apply_click,),
                flat=True,
            )

    @property
    def ui(self):
        if self._ui is None:
            with QLayout(self.server) as layout:
                self._ui = layout

                with quasar.QDrawer(
                    v_model=("leftDrawerOpen", True),
                    side="left",
                    elevated=True,
                ):
                    self.filter_operator_ui()

                    self.filter_options_ui()

                    self.filter_apply_ui()

                with quasar.QPageContainer():
                    with quasar.QPage():
                        with html.Div(classes="row", style="min-height: inherit;"):
                            with html.Div(classes="col q-pa-md"):
                                pass

        return self._ui


def filtering(server=None, *args, **kwargs):
    server = get_server()
    server.client_type = "vue3"

    app = FilteringApp(server)

    a, b, c = (0, 1, 2)

    test_values = [
        [],
        [a],
        [b],
        [c],
        [a, b],
        [a, c],
        [b, c],
        [a, b, c],
    ]

    def on_apply_filter(filter: FilterProtocol):
        result = list(map(filter.evaluate, test_values))

        for value, res in zip(test_values, result):
            print(value, res)

    app.set_on_apply_filter(on_apply_filter)

    app.ui

    server.start(**kwargs)


if __name__ == "__main__":
    filtering()



from typing import Dict

from nrtk_explorer.app.applet import Applet

from trame.widgets import quasar, html
from trame.ui.quasar import QLayout

from trame.app import get_server

from nrtk_explorer.widgets.nrtk_explorer import ParamsWidget

from nrtk_explorer.library.transforms import (
    ImageTransform,
    GaussianBlurTransform,
    IdentityTransform,
    TestTransform,
)


class ParametersApp(Applet):
    def __init__(self, server):
        super().__init__(server)

        self.state.current_transform = "TestTransform"

        self._transforms: Dict[str, ImageTransform] = {
            "TestTransform": TestTransform(),
            "GaussianBlurTransform": GaussianBlurTransform(),
            "IdentityTransform": IdentityTransform(),
        }

        self.state.transforms = [k for k in self._transforms.keys()]
        self.state.current_transform = self.state.transforms[0]

        self.on_apply_transform = lambda: None

        self.server.controller.add("on_server_ready")(self.on_server_ready)

        self._ui = None

    def on_server_ready(self, *args, **kwargs):
        # Bind instance methods to state change
        self.on_current_transform_change()
        self.state.change("current_transform")(self.on_current_transform_change)

    def on_current_transform_change(self, **kwargs):
        print("current transform changed", self.state.current_transform)
        transform = self._transforms[self.state.current_transform]
        self.state.params_values = transform.get_parameters()
        self.state.params_descriptions = transform.get_parameters_description()

    def on_transform_parameters_changed(self, parameters, **kwargs):
        print("on_transform_parameters_changed", parameters, kwargs)
        transform = self._transforms[self.state.current_transform]
        transform.set_parameters(parameters)
        self.state.params_values = transform.get_parameters()

    def transform_select_ui(self):
        with html.Div(trame_server=self.server):
            quasar.QSelect(
                label="Transform",
                v_model=("current_transform",),
                options=(self.state.transforms,),
                filled=True,
                emit_value=True,
                map_options=True,
            )

    def transform_params_ui(self):
        with html.Div(trame_server=self.server):
            ParamsWidget(
                values=("params_values",),
                descriptions=("params_descriptions",),
                valuesChanged=(self.on_transform_parameters_changed, "[$event]"),
            )

    def transform_apply_ui(self):
        with html.Div(trame_server=self.server):
            quasar.QBtn(
                "Apply",
                click=(self.on_apply_transform,),
                classes="full-width",
                flat=True,
            )

    @property
    def ui(self):
        if self._ui is None:
            with QLayout(self.server) as layout:
                self._ui = layout

                with quasar.QDrawer(
                    v_model=("leftDrawerOpen", True),
                    side="left",
                    elevated=True,
                ):
                    self.transform_select_ui()

                    with html.Div(
                        classes="q-pa-md q-ma-md",
                        style="border-style: solid; border-width: thin; border-radius: 0.5rem; border-color: lightgray;",
                    ):
                        self.transform_params_ui()

                    self.transform_apply_ui()

                with quasar.QPageContainer():
                    with quasar.QPage():
                        with html.Div(classes="row", style="min-height: inherit;"):
                            with html.Div(classes="col q-pa-md"):
                                pass

        return self._ui


def main(server=None, *args, **kwargs):
    server = get_server()
    server.client_type = "vue3"

    app = ParametersApp(server)
    app.ui

    server.start(**kwargs)


if __name__ == "__main__":
    main()



from typing import TypeVar, Generic, Dict, TypedDict, Literal, Union, Any, Sequence, Optional

import abc

from PIL import ImageFilter, ImageOps
from PIL.Image import Image

T = TypeVar("T")
S = TypeVar("S")

ParameterType = Literal["string", "integer", "float", "boolean"]
ParameterValue = Union[str, int, float]
ParameterOptions = Sequence[ParameterValue]


# If we target python>=3.11 we should use NotRequired instead of Optional
class ParameterDescription(TypedDict):
    type: ParameterType
    label: str
    description: Optional[str]
    default: Optional[ParameterValue]
    options: Optional[ParameterOptions]


class Transform(abc.ABC, Generic[T, S]):
    @abc.abstractmethod
    def get_parameters(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @abc.abstractmethod
    def set_parameters(self, params: Dict[str, Any]):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        raise NotImplementedError()

    @abc.abstractmethod
    def execute(self, input: T, *input_args: Any) -> S:
        raise NotImplementedError()


class ImageTransform(Transform[Image, Image]):
    pass


class IdentityTransform(ImageTransform):
    def get_parameters(self) -> Dict[str, Any]:
        return {}

    def set_parameters(self, params: Dict[str, Any]):
        pass

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        return {}

    def execute(self, input: Image, *input_args: Any) -> Image:
        return input.copy()


class GaussianBlurTransform(ImageTransform):
    default_radius = 1

    def __init__(self):
        self._radius = GaussianBlurTransform.default_radius

    def get_parameters(self) -> Dict[str, Any]:
        return {"radius": self._radius}

    def set_parameters(self, params: Dict[str, Any]):
        self._radius = params.get("radius", GaussianBlurTransform.default_radius)

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        radius_description: ParameterDescription = {
            "type": "integer",
            "default": GaussianBlurTransform.default_radius,
            "label": "Radius",
            "description": None,
            "options": None,
        }

        return {
            "radius": radius_description,
        }

    def execute(self, input: Image, *input_args: Any) -> Image:
        return input.filter(ImageFilter.GaussianBlur(self._radius))


class InvertTransform(ImageTransform):
    def get_parameters(self) -> dict[str, Any]:
        return {}

    def set_parameters(self, params: Dict[str, Any]):
        pass

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        return {}

    def execute(self, input: Image, *input_args: Any) -> Image:
        return ImageOps.invert(input)


class DownSampleTransform(ImageTransform):
    def get_parameters(self) -> dict[str, Any]:
        return {}

    def set_parameters(self, params: Dict[str, Any]):
        pass

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        return {}

    def execute(self, input: Image, *input_args: Any) -> Image:
        cx = 2
        return input.resize((input.size[0] // cx, input.size[1] // cx))


class TestTransform(ImageTransform):
    default_string = "abc"
    default_int = 7
    default_float = 3.14
    default_boolean = True
    default_select = "two"

    def __init__(self):
        self._string_value = TestTransform.default_string
        self._int_value = TestTransform.default_int
        self._float_value = TestTransform.default_float
        self._boolean_value = TestTransform.default_boolean
        self._select_value = TestTransform.default_select

    def get_parameters(self) -> Dict[str, Any]:
        return {
            "string_param": self._string_value,
            "int_param": self._int_value,
            "float_param": self._float_value,
            "boolean_param": self._boolean_value,
            "select_param": self._select_value,
        }

    def set_parameters(self, params: Dict[str, Any]):
        self._string_value = params.get("string_param", TestTransform.default_string)
        self._int_value = params.get("int_param", TestTransform.default_int)
        self._float_value = params.get("float_param", TestTransform.default_float)
        self._boolean_value = params.get("boolean_param", TestTransform.default_boolean)
        self._select_value = params.get("select_param", TestTransform.default_select)

    def get_parameters_description(self) -> Dict[str, ParameterDescription]:
        string_description: ParameterDescription = {
            "type": "string",
            "default": TestTransform.default_string,
            "label": "A String",
            "description": None,
            "options": None,
        }

        int_description: ParameterDescription = {
            "type": "integer",
            "default": TestTransform.default_int,
            "label": "An Integer",
            "description": None,
            "options": None,
        }

        float_description: ParameterDescription = {
            "type": "float",
            "default": TestTransform.default_float,
            "label": "A Float",
            "description": None,
            "options": None,
        }

        boolean_description: ParameterDescription = {
            "type": "boolean",
            "default": TestTransform.default_boolean,
            "label": "A Boolean",
            "description": None,
            "options": None,
        }

        select_description: ParameterDescription = {
            "type": "string",
            "default": TestTransform.default_select,
            "label": "A Select",
            "description": None,
            "options": ["one", "two", "three", "four"],
        }

        return {
            "string_param": string_description,
            "int_param": int_description,
            "float_param": float_description,
            "boolean_param": boolean_description,
            "select_param": select_description,
        }

    def execute(self, input: Image, *input_args: Any) -> Image:
        return input.copy()



from typing import TypedDict, List


class DatasetCategory(TypedDict):
    id: int
    name: str


class DatasetImage(TypedDict):
    id: int
    file_name: str
    width: int
    height: int


class DatasetAnnotation(TypedDict):
    id: int
    image_id: int
    category_id: int
    bbox: List[int]


class Dataset(TypedDict):
    categories: List[DatasetCategory]
    images: List[DatasetImage]
    annotations: List[DatasetAnnotation]



from trame.widgets import html

from nrtk_explorer.widgets.nrtk_explorer import ImageDetection


def image_list_component(image_ids_key, hover_fn=None, is_transformation=False):
    with html.Div(classes="row"):
        with html.Div(
            v_for=(f"image_id in {image_ids_key}",), key="image_id", classes="col-3 q-pa-sm"
        ):
            ImageDetection(
                identifier=("image_id.match(/\d+/)[0]",),  # noqa: W605
                src=("get(image_id).value",),
                meta=("get(`${image_id}_meta`).value",),
                annotations=("get(`${image_id}_result`).value",),
                categories=("annotation_categories",),
                selected=("(image_id.match(/\d+/)[0] == hovered_id)",),  # noqa: W605
                isTransformation=is_transformation,
                hover=(hover_fn, "[$event]"),
            )



import hashlib

from sklearn.decomposition import PCA


class DimReducerManager:
    def __init__(self):
        self.cached_reducers = {}
        self.cached_reductions = {}

    def reduce(self, name, fit_features, features=None, cache=True, **kwargs):
        fit_features_id = hashlib.md5(fit_features.data).hexdigest()
        reducer_id = (
            fit_features_id + ":" + name + ":" + ":".join("%s=%r" % x for x in kwargs.items())
        )

        reduction_id = reducer_id
        if features is not None:
            features_id = hashlib.md5(features.data).hexdigest()
            reduction_id = reducer_id + ":" + features_id

        if cache is False or reducer_id not in self.cached_reducers:
            reducer = None
            if name.upper() == "PCA":
                reducer = PCAReducer(**kwargs)
            elif name.upper() == "UMAP":
                reducer = UMAPReducer(**kwargs)
            else:
                raise TypeError

            # Train the model
            reducer.fit(fit_features)
            self.cached_reducers[reducer_id] = reducer

        if features is None or len(features) == 0:
            return None
        else:
            if cache is False or reduction_id not in self.cached_reductions:
                # Perform reduction without modifying the model
                self.cached_reductions[reduction_id] = self.cached_reducers[reducer_id].reduce(
                    features
                )
            return self.cached_reductions[reduction_id]


class DimReducer:
    def reduce(self, features):
        raise NotImplementedError

    def fit(self, features):
        raise NotImplementedError


class PCAReducer(DimReducer):
    def __init__(self, dims=3, whiten=False, solver="auto"):
        self._dims = dims
        self._whiten = whiten
        self._solver = solver
        self.pca = PCA(
            n_components=int(self._dims),
            whiten=self._whiten,
            svd_solver=self._solver,
        )

    def reduce(self, features):
        x = self.pca.transform(features)
        return x.tolist()

    def fit(self, features):
        self.pca.fit(features)


class UMAPReducer(DimReducer):
    def __init__(self, dims=3, **kwargs):
        import umap

        self.umap = umap.UMAP(n_components=int(dims), **kwargs)

    def reduce(self, features):
        embeddings = self.umap.transform(features)
        return embeddings.tolist()

    def fit(self, features):
        self.umap.fit(features)



from nrtk_explorer.library import embeddings_extractor
from nrtk_explorer.library import dimension_reducers
from nrtk_explorer.library import images_manager
import nrtk_explorer.test_data

from tabulate import tabulate
from itertools import product

import json
import os
import pytest
import timeit

CURRENT_DIR_NAME = os.path.dirname(nrtk_explorer.test_data.__file__)
DATASET = f"{CURRENT_DIR_NAME}/coco-od-2017/test_val2017.json"


def image_paths_impl():
    with open(DATASET) as f:
        dataset = json.load(f)
    images = dataset["images"]

    paths = list()
    for image_metadata in images:
        paths.append(os.path.join(os.path.dirname(DATASET), image_metadata["file_name"]))
    return paths


@pytest.fixture
def image_paths():
    return image_paths_impl()


def test_features_small(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    assert len(features) == 10
    print(features)


def test_features_zero(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract([])
    assert features is None
    print(features)


@pytest.mark.slow
def test_features_all(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths)
    assert len(features) == len(image_paths)
    print(f"Number of features: {len(features)}")


def test_pca_2d(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    model = dimension_reducers.PCAReducer(2)
    points = model.fit(features)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 2
    print(points)


def test_pca_3d(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    model = dimension_reducers.PCAReducer(3)
    points = model.fit(features)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 3
    print(points)


def test_umap_2d(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    model = dimension_reducers.UMAPReducer(2, n_neighbors=8)
    points = model.fit(features)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 2
    print(points)


def test_umap_3d(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    model = dimension_reducers.UMAPReducer(3, n_neighbors=8)
    points = model.fit(features)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 3
    print(points)


def test_reducer_manager(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths[:10])
    mgr = dimension_reducers.DimReducerManager()
    old_points = mgr.reduce(fit_features=features, features=features, name="PCA", dims=3)
    assert len(old_points) > 0
    assert len(old_points[0]) == 3

    # breakpoint()
    new_points = mgr.reduce(fit_features=features, features=features, name="PCA", dims=3)
    assert id(old_points) == id(new_points)

    new_points_2d = mgr.reduce(fit_features=features, features=features, name="PCA", dims=2)
    assert id(old_points) != id(new_points_2d)


@pytest.mark.slow
def test_features_extractor_benchmark(image_paths):
    repetitions = 3
    sampling = [10, 100]
    batch_size = [1, 8, 16, 32]
    setups = list(product(sampling, batch_size))
    setups.append([500, 1])
    setups.append([500, 64])
    table = list()

    # Pre-load images
    manager = images_manager.ImagesManager()
    for path in image_paths[: max(sampling)]:
        manager.load_image_for_model(path)

    for n, batch_size in setups:
        extractor = embeddings_extractor.EmbeddingsExtractor(manager=manager)
        output = timeit.repeat(
            stmt=lambda: extractor.extract(image_paths[:n], batch_size=batch_size),
            number=repetitions,
            repeat=5,
        )
        table.append([n, batch_size, min(output) / 10 / n])

    print(tabulate(table, headers=["#Samples", "batch_size", "ExecTime(sec)"], tablefmt="github"))


@pytest.mark.slow
def test_reducer_manager_benchmark(image_paths):
    setups = [
        ("PCA", 10, True, 100),
        ("PCA", 10, False, 100),
        ("PCA", 100, True, 10),
        ("PCA", 100, False, 10),
        ("UMAP", 10, True, 100),
        ("UMAP", 10, False, 100),
        ("UMAP", 100, True, 10),
        ("UMAP", 100, False, 10),
    ]

    mgr = dimension_reducers.DimReducerManager()
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths)

    # Short benchmarks cached
    for name, n, cache, iterations in setups:
        output = timeit.timeit(
            lambda: mgr.reduce(
                fit_features=features[:n], features=features, name=name, cache=cache
            ),
            number=iterations,
        )

        print(
            f"{name} dims reductions of {n} images cached={cache} mean(FullExecTime)={output / iterations}s"
        )


@pytest.mark.slow
def test_pca_3d_large(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths)
    model = dimension_reducers.PCAReducer(3)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 3
    print(f"Number of images: {len(points)}")
    print(points)


@pytest.mark.slow
def test_umap_3d_large(image_paths):
    extractor = embeddings_extractor.EmbeddingsExtractor()
    features = extractor.extract(image_paths)
    model = dimension_reducers.UMAPReducer(3)
    points = model.reduce(features)
    assert len(points) > 0
    assert len(points[0]) == 3
    print(f"Number of images: {len(points)}")
    print(points)
